/*           
Module : PStat.cpp
Purpose: Defines the interface for a generalised framework for executing a lengthy operation in a thread. 
         Feedback is provided by a progress dialog which can optionally be cancelled.
Author : PJN / 27-03-1997 Initially created
History: PJN / 18-02-1998 Fixed a bug where the worker thread was trying to close a dialog 
                          before it was created. A CEvent member of CProgressThreadDlg now 
                          protects this. Also removed the usual appwizard comments from the 
                          demo program
         PJN / 8-11-1998  1. Added an option to confirm cancel by means of a message box.
                          2. General tidy up of the code
                          3. Inclusion of VC 5 workspace files now as standard
                          4. All code now compiles cleanly at warning level 4
                          5. Code now supports UNICODE and build configurations are now provided.
         PJN / 28-03-2000 1. Changed dialog resources which pstat uses to use the prefix ID._PSTAT_...
                          2. Update documentation about its usage.
         PJN / 30-03-2000 Fixed a race condition which existing in the worker thread. The issue was 
                          that the dialog could be closed before the CSingleLock destructor was 
                          called in the worker thread which resulted in an access violation.
         PJN / 03-02-2003 1. Reimplemented a busy loop using a CEvent in ProgressDialogWorkerThread.
                          2. Implemented synchronous closure of the worker thread.
                          3. Updated copyright details.
         PJN / 23-07-2007 1. Updated copyright details.
                          2. Optimized CProgressThreadDlg constructor code.
                          3. CProgressThreadInfo is now no longer derived from CObject as it was
                          not really required.
                          4. Code now uses newer C++ style casts instead of C style casts.
                          5. Addition of a CPROGRESSTHREADDLG_EXT_CLASS preprocessor macro to make the 
                          code easier to use in an extension dll.
                          6. Restructured code to avoid need for internal CProgressThreadInfo class.
                          7. Made the Cancel button's ID on the IDD_PSTAT_PROGRESS_WITH_CANCEL "IDCANCEL".
                          8. Code now internally thread protects the "m_bCancelled" and "m_bOkToClose"
                          member variables.
                          9. FUNCTION_WITH_PROGRESS function is now declared as PSTAT_FUNCTION and returns
                          a void. It is also now scoped in the CProgressThreadDlg class.
                          10. Updated sample app to clean compile on VC2005.
                          11. Updated the documentation to use the same style as the web site.
         PJN / 21-01-2017 1. Updated copyright details.
                          2. Updated the code to compile cleanly on VC 2010 - 2015.
                          3. Removed the VC6 style classwizard comments from the code.
                          4. Replaced NULL throughout the codebase with nullptr
                          5. Moved all the configuration values used by the CProgressThreadDlg class into a
                          new CProgressThreadSettings class.
                          6. Added SAL annotations to all the code
                          7. Refactored all the setup code in CProgressThreadDlg::OnInitDialog into various 
                          virtual methods.
                          8. The progress control is now marked as paused when the confirm cancel 
                          messagebox is displayed.
                          9. Updated the sample app's application icon

Copyright (c) 1997 - 2017 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


///////////////////////// Includes ////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"
#include "PStat.h"


///////////////////////// Macros / Defines ////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

#ifndef PBST_NORMAL
#define PBST_NORMAL 0x0001
#endif //#ifndef PBST_NORMAL

#ifndef PBST_PAUSED
#define PBST_PAUSED 0x0003
#endif //#ifndef PBST_PAUSED

#ifndef PBM_SETSTATE
#define PBM_SETSTATE (WM_USER+16)
#endif //#ifndef PBM_SETSTATE


/////////////////////// Implementation ////////////////////////////////////////

IMPLEMENT_DYNAMIC(CProgressThreadDlg, CDialog)

CProgressThreadDlg::CProgressThreadDlg(_In_ FUNCTION_WITH_PROGRESS* pfnFunction, _In_ const CProgressThreadSettings& settings, 
                                       _In_opt_ CWnd* pParent) : CDialog(settings.m_bAllowCancel ? IDD_PSTAT_PROGRESS_WITH_CANCEL : IDD_PSTAT_PROGRESS, pParent),
                                                                 m_pfnFunction(pfnFunction),
                                                                 m_Settings(settings),
                                                                 m_bCancelled(FALSE),
                                                                 m_nCurrentPosition(settings.m_nProgressMinRange),
                                                                 m_bOkToClose(FALSE),
                                                                 m_pThread(nullptr)
{
  m_ftStart.dwHighDateTime = 0;
  m_ftStart.dwLowDateTime = 0;
}

CProgressThreadDlg::~CProgressThreadDlg()
{
  //Shutdown the worker thread if required
  if (m_pThread != nullptr)
  {
    WaitForSingleObject(m_pThread->m_hThread, INFINITE);
    delete m_pThread;
    m_pThread = nullptr;
  }
}

void CProgressThreadDlg::DoDataExchange(CDataExchange* pDX)
{
  //Let the base class do its thing
  __super::DoDataExchange(pDX);

  DDX_Control(pDX, IDC_PROGRESS, m_ctrlProgress);
}

BEGIN_MESSAGE_MAP(CProgressThreadDlg, CDialog)
END_MESSAGE_MAP()

LONG CProgressThreadDlg::HasBeenCancelled()
{
  return InterlockedExchange(&m_bCancelled, m_bCancelled);
}

LONG CProgressThreadDlg::OkToClose()
{
  return InterlockedExchange(&m_bOkToClose, m_bOkToClose);
}

bool CProgressThreadDlg::GetStartTime()
{
  SYSTEMTIME st;
  GetSystemTime(&st);
  if (!SystemTimeToFileTime(&st, &m_ftStart))
    return false;

  return true;
}

bool CProgressThreadDlg::InitializeDialog()
{
  //initialise the progress control
  m_ctrlProgress.SetRange32(m_Settings.m_nProgressMinRange, m_Settings.m_nProgressMaxRange);
  m_ctrlProgress.SetPos(m_nCurrentPosition);

  //Set the caption if required
  if (!m_Settings.m_sCaption.IsEmpty())
    SetWindowText(m_Settings.m_sCaption);
  return true;
}

bool CProgressThreadDlg::CreateWorkerThread()
{
  //Create the worker thread (initally suspended)
  ASSERT(m_pThread == nullptr);
  m_pThread = AfxBeginThread(CProgressThreadDlg::_ProgressDialogWorkerThread, this, m_Settings.m_nThreadPriority, 0, CREATE_SUSPENDED);
  if (m_pThread != nullptr)
  {
    //We are in charge of the deletion of the thread
    m_pThread->m_bAutoDelete = FALSE;

    //Resume the worker thread
    m_pThread->ResumeThread();
  }
  return (m_pThread != nullptr) ? true : false;
}

BOOL CProgressThreadDlg::OnInitDialog()
{
  //Let the parent class do its thing
  __super::OnInitDialog();

  //Call the various virtual methods to get the dialog ready
  if (!GetStartTime())
  {
    TRACE(_T("CProgressThreadDlg::OnInitDialog, Failed to get start time\n"));
    EndDialog(IDCANCEL);
    return TRUE;
  }
  if (!InitializeDialog())
  {
    TRACE(_T("CProgressThreadDlg::OnInitDialog, Failed to initialize dialog\n"));
    EndDialog(IDCANCEL);
    return TRUE;
  }
  if (!CreateWorkerThread())
  {
    TRACE(_T("CProgressThreadDlg::OnInitDialog, Failed to create worker thread\n"));
    EndDialog(IDCANCEL);
    return TRUE;
  }

  return TRUE;
}

void CProgressThreadDlg::OnCancel() 
{
  if (OkToClose())
  {
    //Let the base class do its thing
    __super::OnCancel();
    return;
  }

  if (m_Settings.m_bAllowCancel)
  {
    if (m_Settings.m_bConfirmCancel)
    {
      //Worker thread needs to be suspended while we are displaying the confirm cancel message box
      AFXASSUME(m_pThread != nullptr);
      m_pThread->SuspendThread();

      //Mark the progress control as paused
    #ifdef UNICODE
      m_ctrlProgress.SendMessage(PBM_SETSTATE, PBST_PAUSED, 0L); //Use CWnd::SendMessage rather than CProgressCtrl::SetState as it is declared incorrectly in VS 2013 & 2015
    #endif //#ifdef UNICODE

      //Ask the user if they want to cancel the operation
      if (AfxMessageBox(m_Settings.m_sConfirmPrompt, m_Settings.m_nMessageBoxType) == m_Settings.m_nMessageBoxConfirmation)
        InterlockedExchange(&m_bCancelled, TRUE);

      //Restore the progress control
    #ifdef UNICODE
      m_ctrlProgress.SendMessage(PBM_SETSTATE, PBST_NORMAL, 0L);
    #endif //#ifdef UNICODE

      //Resume the worker thread
      m_pThread->ResumeThread();
    }
    else
      InterlockedExchange(&m_bCancelled, TRUE);
  }
}

void CProgressThreadDlg::OnOK() 
{
  if (OkToClose())
    __super::OnOK();
}

void CProgressThreadDlg::ReportProgress(_In_ int nCurrentPosition) 
{
  //Quick exit?
  if (m_nCurrentPosition == nCurrentPosition)
    return;

  //Update the progress control and cache the current percentage
  m_ctrlProgress.SetPos(nCurrentPosition); 
  m_nCurrentPosition = nCurrentPosition;

  //Work out the relative position
  int nRelativePosition = (nCurrentPosition - m_Settings.m_nProgressMinRange);
  if (nRelativePosition == 0)
    return;

  //Work out the time left based on the current time and current percentage done
  SYSTEMTIME st;
  GetSystemTime(&st);
  FILETIME ft;
  if (!SystemTimeToFileTime(&st, &ft))
    return;
  __int64 nMillisecondsUsed = ((static_cast<__int64>(ft.dwHighDateTime) << 32) + ft.dwLowDateTime - (static_cast<__int64>(m_ftStart.dwHighDateTime) << 32) - m_ftStart.dwLowDateTime) / 10000;
  __int64 nMillisecondsLeft = (nMillisecondsUsed / nRelativePosition) * ((m_Settings.m_nProgressMaxRange - m_Settings.m_nProgressMinRange) - nRelativePosition);
  DWORD dwMinutes = static_cast<DWORD>(nMillisecondsLeft / 60000);
  DWORD dwSeconds = static_cast<DWORD>((nMillisecondsLeft/1000) % 60);

  //Construct the time left text
  CString sText;
  if (dwMinutes)
  {
    CString sVal;
    sVal.Format(_T("%u"), dwMinutes);
    AfxFormatString1(sText, IDS_PSTAT_TIMELEFT_IN_MINUTES, sVal);
  }
  else
  {
    if (dwSeconds)
    { 
      CString sVal;
      sVal.Format(_T("%u"), dwSeconds);
      AfxFormatString1(sText, IDS_PSTAT_TIMELEFT_IN_SECONDS, sVal);
    }
    else 
    {
      //if 0 seconds are left then display no text
      sText.Empty();
    }
  }

  //Update the UI text
  ASSERT(GetDlgItem(IDC_TIMELEFT));
  GetDlgItem(IDC_TIMELEFT)->SetWindowText(sText);
}

UINT CProgressThreadDlg::_ProgressDialogWorkerThread(_In_ LPVOID pParam)
{
  //Convert from the SDK world to the C++ world
  CProgressThreadDlg* pDlg = static_cast<CProgressThreadDlg*>(pParam);
  AFXASSUME(pDlg != nullptr);
  return pDlg->ProgressDialogWorkerThread();
}

UINT CProgressThreadDlg::ProgressDialogWorkerThread()
{
  //Call the function
  ASSERT(m_pfnFunction != nullptr);
  m_pfnFunction(this, m_Settings.m_dwItemData);

  //close down the progress dialog with the correct result
  InterlockedExchange(&m_bOkToClose, TRUE);
  if (HasBeenCancelled())
    PostMessage(WM_COMMAND, IDCANCEL);
  else
    PostMessage(WM_COMMAND, IDOK);

  return 0;
}
